import io
import json
from dataclasses import dataclass
from functools import cached_property
from typing import Any, Dict, List


class Node:
    def __hash__(self):
        return id(self)


@dataclass
class ListNode(Node):
    children: List[Node]

    @property
    def section(self):
        return self.children and all(child.section for child in self.children)

    @cached_property
    def length(self):

        child_lengths = sum(child.length for child in self.children)
        if len(self.children) == 0:
            return 2
        else:
            return child_lengths + 2 * (len(self.children) - 1) + 4


@dataclass
class Pair:
    key: str
    value: Node

    @property
    def section(self):
        return self.value.section


@dataclass
class DictNode(Node):
    children: List[Pair]
    section: bool = True

    @cached_property
    def length(self):
        child_lengths = sum(
            len(pair.key) + 3 + pair.value.length for pair in self.children
        )
        if len(self.children) == 0:
            return 2
        else:
            return child_lengths + 2 * (len(self.children) - 1) + 4


@dataclass
class LeafNode(Node):
    value: Any
    section = False

    @cached_property
    def length(self):
        return len(json.dumps(self.value))


def parse_tree(data):
    if isinstance(data, dict):
        return DictNode([Pair(k, parse_tree(v)) for k, v in data.items()])
    elif isinstance(data, list):
        return ListNode([parse_tree(v) for v in data])
    else:
        return LeafNode(data)


def write_horizontal_list(obj, output, indent, max_width):
    output.write("[ ")
    for i, child in enumerate(obj.children):
        if i != 0:
            output.write(", ")
        write_compound(0, child, output, indent, max_width)
    output.write(" ]")


def write_horizontal_dict(obj, output, indent, max_width):
    output.write("{ ")
    for i, pair in enumerate(obj.children):
        key, child = pair.key, pair.value
        if child != LeafNode(None):
            if i != 0:
                output.write(", ")
            key_str = f"{key} = "
            output.write(key_str)
            write_compound(0, child, output, indent, max_width)
    output.write(" }")


def write_vertical_list(current_indent, obj, output, indent, max_width):
    output.write("[\n")
    for i, child in enumerate(obj.children):
        child_indent = current_indent + indent
        output.write(" " * child_indent)
        write_compound(child_indent, child, output, indent, max_width)
        if i != len(obj.children):
            output.write(", ")

        output.write("\n")

    output.write(" " * current_indent)
    output.write("]")


def write_vertical_dict(current_indent, obj, output, indent, max_width):
    output.write("{\n")
    for i, pair in enumerate(obj.children):
        key, child = pair.key, pair.value
        if child != LeafNode(None):
            if i != 0:
                output.write(",\n")
            child_indent = current_indent + indent
            output.write(" " * child_indent)
            key_str = f"{key} = "
            output.write(key_str)
            write_compound(child_indent, child, output, indent, max_width)
    output.write("\n")
    output.write(" " * current_indent)
    output.write("}")


def write_compound(current_indent, obj, output, indent, max_width):

    vertical = current_indent + obj.length > max_width

    if isinstance(obj, ListNode) and vertical:
        write_vertical_list(current_indent, obj, output, indent, max_width)
    elif isinstance(obj, ListNode) and not vertical:
        write_horizontal_list(obj, output, indent, max_width)
    elif isinstance(obj, DictNode) and vertical:
        write_vertical_dict(current_indent, obj, output, indent, max_width)
    elif isinstance(obj, DictNode) and not vertical:
        write_horizontal_dict(obj, output, indent, max_width)
    else:
        write_leaf(obj, output)


def write_child(current_indent, child, output, indent, max_width):
    if isinstance(child, LeafNode):
        write_leaf(child, output)
        output.write("\n")
    else:
        write_compound(current_indent, child, output, indent, max_width)
        output.write("\n")


def write_leaf(leaf, output):
    if isinstance(leaf.value, str) and len(leaf.value) > 255:
        output.write('"""')
        output.write(leaf.value)
        output.write('"""')
    else:
        output.write(json.dumps(leaf.value))


def write_section(section, output, indent, max_width):

    for pair in section.children:
        key, child = pair.key, pair.value
        if child != LeafNode(None):
            output.write(f"{key} = ")
            write_child(0, child, output, indent, max_width)
    output.write("\n")


def dump(data, f, indent=4, max_width=100):
    output = f
    tree = parse_tree(data)

    blank_section = DictNode([])
    sections = [("", blank_section)]
    for pair in tree.children:
        key, value = pair.key, pair.value
        if value.section:
            if isinstance(value, DictNode):
                sections.append((f"[{key}]", value))
            elif isinstance(value, ListNode):
                for child in value.children:
                    sections.append((f"[[{key}]]", child))
        else:
            blank_section.children.append(Pair(key, value))

    for key, section in sections:
        if key:
            output.write(key)
            output.write("\n")
        write_section(section, output, indent, max_width)


def dumps(data, indent=4, max_width=100):
    f = io.StringIO()
    dump(data, f, indent, max_width)
    return f.getvalue()



import time
import sys


def deep_get(obj, path):
    if not path:
        return obj
    elif isinstance(obj, dict):
        return deep_get(obj[path[0]], path[1:])
    elif isinstance(obj, list):
        return [deep_get(child, path) for child in obj]
    else:
        return obj



import requests
import tempfile
import boto3
from urllib.parse import urlparse
from time import sleep
from .relycomply_gql_client import RelyComplyGQLClient
from .exceptions import RelyComplyClientException

s3_client = boto3.client("s3")


class RelyComplyClient:
    def __init__(self, **kwargs):

        # Expose the raw-ish gql client
        self.gql = RelyComplyGQLClient(**kwargs)

    def object_id(self, obj_or_id):
        """
        Useful helpers so we can pass json objects with and "id" field instead of
        just ids
        """
        if isinstance(obj_or_id, dict):
            return obj_or_id["id"]

    def object_ids(self, objs_or_ids):
        if not objs_or_ids:
            return []
        return [self.object_id(obj) for obj in objs_or_ids]

    def sign_url(self, url):
        """
        If an s3 url is passed this will sign it, else if https just return
        """

        if url.startswith("https://"):
            return url

        if url.startswith("s3://"):
            # pull out the bucket and path
            o = urlparse(url)
            return s3_client.generate_presigned_url(
                "get_object",
                Params={
                    "Bucket": o.netloc,
                    "Key": o.path.lstrip("/"),
                },
                ExpiresIn=5 * 60,
            )

    def put_to_datafile(self, df, name):
        """
        Takes a dataframe converst it to parquet, creates a new datafile, and puts it at the putUrl
        """
        datafile = self.gql.createDataFile(name=name)

        with tempfile.NamedTemporaryFile() as fp:
            df.to_parquet(
                fp.name,
                allow_truncated_timestamps=True,
                use_deprecated_int96_timestamps=True,
            )
            try:
                requests.put(datafile["putUrl"], data=fp).raise_for_status()
            except Exception as e:
                raise RelyComplyClientException(
                    f"Error uploading to DataFile {datafile['id']}:{str(e)}"
                )

        return datafile

    def pull_to_datafile(self, source_url, name, wait_for_ready=False):
        signed_url = self.sign_url(source_url)
        datafile = self.gql.createDataFile(name=name, sourceUrl=signed_url)

        if wait_for_ready:
            while True:  # Should have timeout
                sleep(5)
                datafile = self.gql.dataFiles(id=self.object_id(datafile), _only=True)
                if datafile["status"] != "LOADING":
                    break
        if datafile["status"] == "ERROR":
            raise RelyComplyClientException(
                f"Error pulling datafile {datafile['id']}: {datafile['error']}"
            )
        return datafile

    def ingest_datasource(
        self, datasource_name, data_file, format="PARQUET", raw_data_files=None
    ):
        """
        Ingests a datasource
        """
        version = self.gql.ingestDataSource(
            id=datasource_name,
            dataFile=self.object_id(data_file),
            rawDataFiles=self.object_ids(raw_data_files),
            format=format,
        )
        status = version["status"]
        while status == "UPLOADING":
            sleep(5)
            version = self.gql.dataSourceVersions(id=version["id"], _only=True)
            status = version["status"]

        if status == "ERROR":
            raise RelyComplyClientException(
                f"Error ingesting DataSourceVersion {version['id']}: {version['error']}"
            )

        return version

    def run_monitor(self, monitor_name, source_versions):
        """
        Runs a monitor
        """
        return self.gql.runMonitor(
            id=monitor_name, versions=self.object_ids(source_versions), engine="athena"
        )



from gql import Client
from gql.transport.requests import RequestsHTTPTransport
import requests

from relycomply_client.exceptions import RelyComplyClientException
from .gql_generator import GQLGenerator
from gql import gql
from gql.transport.exceptions import TransportProtocolError, TransportQueryError
from .credential_sources import StandardCredentials
from littleutils import only
import logging

log = logging.getLogger(__name__)


class RelyComplyGQLClient:
    """
    A flexible and intelligent GraphQL client for RelyComply. This client will create methods
    that match the mutation sand queries of the RelyComply API, and expose them with familiar
    calling conventions. It also handles paging as well as simplifying the returned structures.

    Queries can be called with their lowerCase field name and any filter arguments as kwargs, e.g.:

        client.products(nameContain="ZA") # Will return a list of products
        client.products(nameContain="ZA", _iter=True) # Will return a lazy generator
        client.products(name="retailZA", _only=True) # Will return only the first object or None

    Mutations can be called in a similar way, but arguments will be lifted into the $input variable

        client.createProduct(name="retailZA", label="South African Retail) # Returns the created product

    The interface is automatically generated from the GQL schema as well as the CLI support templates.
    Thus it should always be in sync.

    """

    def __init__(self, **kwargs):
        credentials = StandardCredentials(**kwargs)

        log.debug(
            f"Credentials: {', '.join([f'{k}={v}' for k,v in credentials.credentials().items()])}"
        )
        # Normalise the url to deal with older config files
        url = credentials.get("url")
        url = url.rstrip("/").rstrip("/graphql")

        # Setup a GQL client, pull creds from env
        transport = RequestsHTTPTransport(
            url=url + "/graphql/",
            verify=True,
            retries=3,
            headers={
                "IMPERSONATE": credentials.get("impersonate"),
                "Authorization": f"Bearer {credentials.require('token')}",
            },
        )
        try:
            self.client = Client(transport=transport, fetch_schema_from_transport=True)
        except TransportProtocolError:
            raise RelyComplyClientException(
                f"Could not retrieve schema from the given url: {url}"
            )
        except requests.exceptions.ConnectionError:
            raise RelyComplyClientException(
                f"Could not connect to the given url: {url}"
            )
        except TypeError:
            raise RelyComplyClientException(
                "The server could not authenticate you, are you sure your token is correct."
            )

        # Pull support templates
        self.templates = requests.get(url + "/cli/v1/support_templates/").json()

        self.schema = self.client.schema

        self.generator = GQLGenerator(self.schema, self.templates)
        self.mutation_fields = []
        self.query_fields = []

        self._attach_mutations()
        self._attach_queries()

    def _attach_mutations(self):
        for field in self.schema.mutation_type.fields.keys():
            self._attach_mutation(field)

    def _attach_mutation(self, field):
        def mutation(**variables):
            return self.execute_mutation(field, variables)

        self.mutation_fields.append(field)
        setattr(self, field, mutation)

    def _attach_queries(self):
        for field in self.schema.query_type.fields.keys():
            self._attach_query(field)

    def _attach_query(self, field):
        def query(_iter=False, _only=False, **variables):
            return self.execute_query(field, variables, _iter, _only)

        self.query_fields.append(field)
        setattr(self, field, query)

    def _collapse_edges(self, value):

        if isinstance(value, list):
            return [self._collapse_edges(child) for child in value]

        if isinstance(value, dict):
            replacement = {}
            for key, child in value.items():
                if (
                    isinstance(child, dict)
                    and len(child) == 1
                    and "edges" in child
                    and isinstance(child["edges"], list)
                ):
                    replacement[key] = [
                        self._collapse_edges(edge["node"]) for edge in child["edges"]
                    ]
                else:
                    replacement[key] = self._collapse_edges(child)
            return replacement
        else:
            return value

    def _only_value(self, result):
        return only(result.values())

    def _execute(self, query, variables):
        result = self.client.execute(query, variable_values=variables)
        return self._only_value(result)

    def _endpoint_graphql(self, action, endpoint):
        return self.generator.generate_endpoint(action, endpoint)

    def _endpoint_table_graphql(self, action, endpoint):
        return self.generator.generate_table_endpoint(action, endpoint)

    def _execute_endpoint(self, action, endpoint, variables, table_mode=False):
        if table_mode:
            graphql = self._endpoint_table_graphql(action, endpoint)
        else:
            graphql = self._endpoint_graphql(action, endpoint)

        query = gql(graphql)
        try:
            return self._execute(query, variables)
        except TransportQueryError as e:
            errors = e.errors
            raise RelyComplyClientException(str(errors[0]["message"])) from e

    def execute_query(self, endpoint, variables, _iter=False, _only=False):
        if _only:
            try:
                return next(self._execute_query_iter(endpoint, variables))
            except StopIteration:
                return None

        if _iter:
            return self._execute_query_iter(endpoint, variables)
        else:
            return list(self._execute_query_iter(endpoint, variables))

    def execute_query_for_table(self, endpoint, variables):
        return self.generator.filter_list_fields(
            endpoint, self._execute_query_iter(endpoint, variables, table_mode=True)
        )

    def _execute_query_iter(self, endpoint, variables, table_mode=False):
        after = None
        hasNextPage = True

        while hasNextPage:
            result = self._execute_endpoint(
                "query", endpoint, variables, table_mode=table_mode
            )
            hasNextPage = result["pageInfo"]["hasNextPage"]

            for edge in result["edges"]:
                node = edge["node"]
                yield self._collapse_edges(node)

            if not hasNextPage:
                break
            else:
                after = result["pageInfo"]["endCursor"]
                variables["after"] = after

    def execute_mutation(self, endpoint, variables):
        result = self._execute_endpoint("mutation", endpoint, dict(input=variables))
        return self._only_value(self._collapse_edges(result))

    def graphql(self, graphql_query, variables, **kwargs):
        """
        Execute a raw graphql query with the underlying client. No postprocessing will be done.

        This is useful because authentication credentials will be sorted out. kwargs will be passed
        through to the execute call of the raw client.
        """
        query = gql(graphql_query)
        return self.client.execute(query, variable_values=variables, **kwargs)



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['relycomply_client']

package_data = \
{'': ['*']}

install_requires = \
['PyYAML',
 'Pygments',
 'boto3',
 'gql==3.0.0rc0',
 'littleutils',
 'pandas',
 'pyarrow',
 'requests',
 'requests-toolbelt',
 'tabulate',
 'termcolor',
 'toml']

entry_points = \
{'console_scripts': ['rely = relycomply_client.cli:main']}

setup_kwargs = {
    'name': 'relycomply-client',
    'version': '0.8.1',
    'description': 'A python client for the RelyComply platform',
    'long_description': '# RelyComply Python Client and CLI\n\n**BETA RELEASE**\n\nThis package contains the python client and CLI for the RelyComply platform:\n\n> RelyComply is an end-to-end Anti-Money Laundering (AML) Plaftorm, managing detection, risk management and automation of your AML compliance requirements\n\nThe CLI makes configuration of the system substantially simpler and allows for a full configuration-as-devops experience. \n\nThe python client exposes both a lower level GraphQL client which makes it easy to interact with the GraphQL APi in a pythonic manner. As well as a higher-level integration client that provides useful routines for common integration tasks.\n\n## Credentials\n\nCredentials for the various clients can be loaded in multiple ways. In the following order of precedence:\n\n* Constructor Arguments\n* Environment Variables\n* AWS Secrets\n* Configuration Files (.rely.toml)\n* Defaults\n\nCredentials can stack, so you can define certain credentials in a cofig file for example, and override it with an environment variable.\n\nThe following credentials can be set:\n\n* token: The API token of the user\n* url: The url of the RelyComply application (default: `https://app.relycomply.com`)\n* impersonate: If user impersonation is available for your user this allows you to make a command on behalf of another user. This should be in format `<organisation_name>:<user_email>`, for example `relycomply:james@relycomply.com`.\n\n**Constructor Arguments**\n\nYou can pass credentials when constructing a client as keyword arguments, e.g.\n\n```python\nclient = RelyComplyGQLClient(token="<token>")\n```\n\n**Environment Variables**\n\nThe client will look if there are matching environment variables of the format `RELYCOMPLY_<CREDENTIAL>`, for example: `RELYCOMPLY_TOKEN=<token>`.\n\n**AWS Secrets**\n\nThe client can integrate with the AWS Secrets Manager. This is done by setting appropriate environment variables with the naming convention `RELYCOMPLY_<CREDENTIAL>_AWS_SECRET` and the value being the secrets path, for example:\n\n```\nRELYCOMPLY_TOKEN_AWS_SECRET=path/to/my/secret\n```\n\nFor more information please consult the [AWS Secrets Manager documentation](https://docs.aws.amazon.com/secretsmanager/).\n\n**Config Files**\n\nCredentials can be set in a `.rely.toml` file. The client will search the current working directory, and all parent directories in order, with the most local directory taking precedence. The `.rely.toml` file is a TOML file with the credentials as keys, e.g.\n\n```toml\ntoken="<token>"\nurl="https://relycomply.customer.com"\n```\n\n**Default Credentials**\n\nThe following credentials have default values:\n\n```toml\nurl="https://app.relycomply.com"\n```\n\n## RelyComplyGQLClient\n\nA flexible and intelligent GraphQL client for RelyComply. This client will create methods that match the mutation sand queries of the RelyComply API, and expose them with familiar calling conventions. It also handles paging as well as simplifying the returned structures.\n\nIt can be constructed as below:\n\n```python\nfrom relycomply_client import RelyComplyGQLClient\n\nclient = RelyComplyGQLClient()\n\n# Or with specific credentials\nclient = RelyComplyGQLClient(token="<token>")\n```\n\nQueries can be called with their lowerCase field name and any filter arguments as kwargs, e.g.:\n\n```python\nclient.products(nameContain="ZA") # Will return a list of products\nclient.products(nameContain="ZA", _iter=True) # Will return a lazy generator\nclient.products(name="retailZA", _only=True) # Will return only the first object or None\n```\n\nThe client will automatically collapse edge lists into plain lists of objects to make the output easier to work with.\n\nMutations can be called in a similar way, but arguments will be lifted into the $input variable\n\n```python\nclient.createProduct(name="retailZA", label="South African Retail") # Returns the created product\n```\n\nThe interface is automatically generated from the GQL schema as well as the CLI support templates. Thus it should always be in sync with the latest features on the platform.\n\nThe client also exposes a raw GraphQl call when you need to make a more complex query. No post processing will be done on the results. For example to query the first 10 products.\n\n```python\nclient.graphql(\n    """\n    products(first:$first) {\n        edges {\n            node {\n                id\n                name\n            }\n        }\n    }\n    """, \n    variables=dict(first=10)\n)\n```\n\n## RelyComplyClient\n\nThe RelyComplyCLient contains higher level methods that make common integration tasks simpler. It provides simple integration with various cloud services and common data tools like pandas.\n\n```python\nfrom relycomply_client import RelyComplyGQLClient\nrc = RelyComplyClient()\n\n# Or for quick usage for a standard client\nfrom relycomply_client.rc import rc\n```\n\nA quick overview of a common data integration with the transaction monitoring is shown below:\n\n```python\n# Load a file with pandas\nraw_df = pd.read_csv(file_path)\n\n# Perform some cleaning\ndf = clean_raw_df(raw_df)\n\n# Pull in a datafile from a known source. \n# This will automatically creat a signed URL if an S3 path is passed\nraw_data_file = rc.pull_to_datafile(\n    file_path, "raw/" + file_name, wait_for_ready=True\n)\n\n# Upload a dataframe as parquet datafile\nprocessed_data_file = rc.put_to_datafile(df, "processed/" + file_name)\n\n# Ingest the given files\n\n# Note that the responses from previous calls can be passed as is for the call\n# arguments. Their id\'s will be automatically extracted.\ndata_source_version = rc.ingest_datasource(\n    data_source, data_file=processed_data_file, raw_data_files=[raw_data_file]\n)\n\n# Run a monitor\nrc.run_monitor(monitor_name, source_versions=[data_source_version])\n```\n\nThe underlying `RelyComplyGQLClient` can be accessed with `.gql` property.\n\n```python\nrc.gql.createProduct(name="bank_account", label="My Bank Account)\n```\n\n## Command Line Interface (CLI)\n\nThe command line interface is an important part of our developer first mentality. It acts as a layer on top of the GraphQL API and makes it substantially easier to for power-users to explore and manipulate RelyComply.\n\nGraphQL is excellent as an API for integration, but can be a lot extra overhead to quickly just see what is happening in the system. Primarily this is because the user has to define the output format they want. This greatly improves the flexibility but certainly is not as easy as just using curl on a rest endpoint. The rely CLI makes it easy to perform the standard queries and mutations on the GraphQL API, without extra effort by the user.\n\nThe CLI can be accessed with the `rely` command. The basic format is to call it with a `type` and an `action`. This will automatically be coerced into the appropriate GraphQL calls. \n\nArguments can be passed as keyword arguments of the form `--key=value` additionally a configuration file name can be passed as the final argument. The value can be a json string, which will be parsed correctly for complex arguments. Arguments are merged with the command line arguments taking precedence.\n\nQueries can be performed with the `list` and `retrieve` actions, `retrieve` will return a single item, and `list` will display a table of items.\n\n```bash\nrely product list # Will list all the products\nrely product list --nameContains="za" # Will list all the products with za in their name\nrely product retrieve --id=123 # Will return just the specified product \n```\n\nMutations can called by their name, with the action being prepended to the type. The system is intelligent enough that the case of the action and type do not matter.\n\n```bash\n# Will call createProduct\nrely product create --name="bank_account" --label="Bank Account" \n\n# Will update the given product (updateProduct) based on the given ID and the config file (pr_my_product.toml)\nrely product update --id=10 pr_my_product.toml\n```\n\nCertain aliases are provided for convenience, e.g. \n\n```bash\n# This will call addCaseNote\nrely case addNote --case=123 --note="This is my note"\n```\n\nThe format of the output can be controlled with the `--json`, `--yaml` and `--toml` (default) flags.',
    'author': 'James Saunders',
    'author_email': 'james@relycomply.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://www.relycomply.com',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.8,<4',
}


setup(**setup_kwargs)
